from typing import Any, Union
from django.utils.functional import cached_property

from .base import BaseEditorJSFeature

import bs4


class InlineEditorJSFeature(BaseEditorJSFeature):
    must_have_attrs = None
    can_have_attrs = None

    def __init__(
        self,
        tool_name: str,
        tag_name: str,
        klass: str = None,
        must_have_attrs: dict = None,
        can_have_attrs: dict = None,
        js: Union[str, list[str]] = None,
        css: Union[str, list[str]] = None,
        include_template: str = None,
        config: dict = None,
        weight: int = 0,
        allowed_tags: list[str] = None,
        allowed_attributes: dict[str, list[str]] = None,
        **kwargs,
    ):
        super().__init__(
            tool_name,
            klass,
            js,
            css,
            include_template,
            config,
            weight=weight,
            allowed_tags=allowed_tags,
            allowed_attributes=allowed_attributes,
            **kwargs,
        )

        must_have_attrs = must_have_attrs or {}
        can_have_attrs = can_have_attrs or {}

        if self.must_have_attrs:
            must_have_attrs.update(self.must_have_attrs)

        if self.can_have_attrs:
            can_have_attrs.update(self.can_have_attrs)

        self.tag_name = tag_name
        self.must_have_attrs = must_have_attrs
        self.can_have_attrs = can_have_attrs

    def build_elements(self, inline_data: list, context: dict[str, Any] = None) -> list:
        """
        Builds the elements for the inline data.

        See the LinkFeature class for an example.
        """
        pass

    def filter(self, item):
        """
        Filter function for the bs4 find_all method.
        Extracts the tag name and attributes from the item and compares them to the must_have_attrs.
        If the item attributes matches the must_have_attrs the element is what we need.
        """
        if item.name != self.tag_name:
            return False

        for key, value in self.must_have_attrs.items():
            if not value:
                if not item.has_attr(key):
                    return False
            else:
                cmp = item.get(key)
                if isinstance(cmp, str):
                    cmp = cmp.strip()
                    if cmp != value:
                        return False
                elif isinstance(cmp, list):
                    if value not in cmp:
                        return False
                else:
                    return False

        return True

    def parse_inline_data(self, soup: bs4.BeautifulSoup, context=None):
        """
        Finds inline elements by the must_have_attrs and can_have_attrs.
        Designed to be database-efficient; allowing for gathering of all data before
        making a database request.

        I.E. For a link; this would gather all page ID's and fetch them in a single query.
        """

        matches: dict[Any, dict[str, Any]] = {}
        elements = soup.find_all(self.filter)
        if not elements:
            return None

        for item in elements:
            matches[item] = {}

            for key in self.must_have_attrs.keys():
                matches[item][key] = item.get(key)

            for key, value in self.can_have_attrs.items():
                if value:
                    matches[item][key] = item.get(key)
                else:
                    if item.has_attr(key):
                        matches[item][key] = True

        # Build all inlines.
        self.build_elements(list(matches.items()), context=context)

    @classmethod
    def get_test_data(cls) -> list[tuple[str, str]]:
        """
        Returns a list of test data.

        The test data should be a list of tuples.

        The first item in the tuple is the raw HTML tag.
        The second item is the expected output.

        The raw HTML tag(s) will be randomly appended into a soup.
        We will use assertQueries(1) to ensure any database queries are kept to a minimum.
        """
        return []


class ModelInlineEditorJSFeature(InlineEditorJSFeature):
    model = None
    chooser_class = None
    tag_name = "a"
    id_attr = "data-id"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tag_name=self.tag_name, **kwargs)
        self.must_have_attrs = self.must_have_attrs | {
            self.id_attr: None,
            f"data-{self.model._meta.model_name}": None,
        }

    @cached_property
    def widget(self):
        if self.chooser_class is None:
            return None
        
        return self.chooser_class()

    def get_id(self, item, attrs: dict[str, Any], context: dict[str, Any] = None):
        return int(attrs[self.id_attr])

    def get_config(self, context: dict[str, Any]):
        config = super().get_config() or {}
        config.setdefault("config", {})
        config["config"][
            "chooserId"
        ] = f"editorjs-{self.model._meta.model_name}-chooser-{context['widget']['attrs']['id']}"
        return config

    def render_template(self, context: dict[str, Any] = None):
        if not self.widget:
            return super().render_template(context)
        
        return self.widget.render_html(
            f"editorjs-{self.model._meta.model_name}-chooser-{context['widget']['attrs']['id']}",
            None,
            {
                "id": f"editorjs-{self.model._meta.model_name}-chooser-{context['widget']['attrs']['id']}"
            },
        )

    def build_element(self, item, obj, context: dict[str, Any] = None):
        """
        Build the element from the object.

        item:    bs4.element.Tag
        obj:     Model
        context: RequestContext | None
        """
        # delete all attributes
        for key in list(item.attrs.keys()):
            del item[key]

        request = None
        if context:
            request = context.get("request")
            item["href"] = self.get_full_url(obj, request)
        else:
            item["href"] = self.get_url(obj)
        item["class"] = f"{self.model._meta.model_name}-link"

    @classmethod
    def get_url(cls, instance):
        return instance.url

    @classmethod
    def get_full_url(cls, instance, request):
        return request.build_absolute_uri(cls.get_url(instance))

    def build_elements(self, inline_data: list, context: dict[str, Any] = None) -> list:
        """
        Process the bulk data; fetch all pages in one go
        and build the elements.
        """
        super().build_elements(inline_data, context=context)
        ids = []
        # element_soups = []
        for data in inline_data:
            # soup: BeautifulSoup
            # element: EditorJSElement
            # matches: dict[bs4.elementType, dict[str, Any]]
            # data: dict[str, Any] # Block data.
            item, data = data

            # # Store element and soup for later replacement of content.
            # element_soups.append((soup, element))

            # Item is bs4 tag, attrs are must_have_attrs

            id = self.get_id(item, data, context)
            ids.append((item, id))

            # delete all attributes
            for key in list(item.attrs.keys()):
                del item[key]

        # Fetch all objects
        objects = self.model.objects.in_bulk([id for item, id in ids])
        for item, id in ids:
            self.build_element(item, objects[id], context)

    def get_css(self):
        return self.widget.media._css.get("all", []) + super().get_css()

    def get_js(self):
        return (self.widget.media._js or []) + super().get_js()

    @classmethod
    def get_test_queryset(cls):
        return cls.model.objects.all()

    @classmethod
    def get_test_data(cls):
        # This test only works for the default build_element method.
        # Any extra attributes will make this test fail otherwise.
        if cls.build_element is not ModelInlineEditorJSFeature.build_element:
            raise ValueError(
                "You must implement the get_test_data method for your ModelInlineEditorJSFeature"
                "if you override the build_element method."
            )

        # Limit test QS to 5.
        models = cls.get_test_queryset()[0:5]
        return [
            (
                f"<a data-id='{model.id}' data-{cls.model._meta.model_name}='True'></a>",
                f"<a href='{cls.get_url(model)}' class='{cls.model._meta.model_name}-link'></a>",
            )
            for model in models
        ]



from typing import Any, TYPE_CHECKING
from django import forms
from django.urls import reverse, path
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.http import (
    HttpRequest,
    FileResponse,
    HttpResponse,
)

from wagtail.images.widgets import AdminImageChooser
from wagtail.images import get_image_model
from wagtail.images.models import SourceImageIOError

if TYPE_CHECKING:
    from wagtail.images.models import Image as AbstractImage


from ..registry import (
    EditorJSFeature,
    EditorJSBlock,
    EditorJSElement,
    FeatureViewMixin,
    wrapper,
)

Image = get_image_model()


class BaseImageFeature(FeatureViewMixin, EditorJSFeature):

    def get_urlpatterns(self):
        return [
            path(
                f"{self.tool_name}/",
                self.handler,
                name=f"{self.tool_name}_for_id_fmt",
            ),
            path(
                f"{self.tool_name}/<int:image_id>/",
                self.handler,
                name=self.tool_name,
            )
        ]

    def handle_get(self, request: HttpRequest, image_id: int) -> "AbstractImage":
        image: "AbstractImage" = get_object_or_404(Image, pk=image_id)

        # Get/generate the rendition
        try:
            rendition = image.get_rendition("original")
        except SourceImageIOError:
            return HttpResponse(
                "Source image file not found", content_type="text/plain", status=410
            )

        with rendition.get_willow_image() as willow_image:
            mime_type = willow_image.mime_type

        # Serve the file
        f = rendition.file.open("rb")
        response = FileResponse(f, filename=image.filename, content_type=mime_type)
        response["Content-Length"] = f.size
        return response


class ImageFeature(BaseImageFeature):
    allowed_tags = ["img", "figure", "figcaption"]
    allowed_attributes = {
        "img": ["src", "alt", "class", "style"],
        "figure": ["class", "style"],
        "figcaption": ["class"],
    }
    klass = "WagtailImageTool"
    js = [
        "wagtail_editorjs/js/tools/wagtail-image.js",
    ]

    def get_config(self, context: dict[str, Any]):
        config = super().get_config() or {}
        config.setdefault("config", {})
        config["config"]["imageChooserId"] =\
            f"editorjs-image-chooser-{context['widget']['attrs']['id']}"
        config["config"]["getImageUrl"] = reverse(f"wagtail_editorjs:{self.tool_name}_for_id_fmt")
        return config

    def validate(self, data: Any):
        super().validate(data)

        d = data["data"]
        if "imageId" not in d:
            raise forms.ValidationError("Invalid imageId value")
            
        
    def render_template(self, context: dict[str, Any] = None):
        widget_id = f"editorjs-image-chooser-{context['widget']['attrs']['id']}"
        return AdminImageChooser().render_html(
            widget_id,
            None,
            {"id": widget_id}
        )
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        image = block["data"].get("imageId")
        image = Image.objects.get(id=image)

        classlist = []
        styles = {}
        if block["data"].get("withBorder"):
            classlist.append("with-border")

        if block["data"].get("stretched"):
            classlist.append("stretched")

        if block["data"].get("backgroundColor"):
            styles["background-color"] = block["data"]["backgroundColor"]
            classlist.append("with-background")

        attrs = {}
        if classlist:
            attrs["class"] = classlist

        if styles:
            attrs["style"] = styles

        url = image.file.url
        if not any([url.startswith(i) for i in ["http://", "https://", "//"]]) and context:
            request = context.get("request")
            if request:
                url = request.build_absolute_uri(url)

        # Caption last - we are wrapping the image in a figure tag
                
        imgTag = EditorJSElement(
            "img",
            close_tag=False,
            attrs={
                "src": url,
                "alt": block["data"].get("alt"),
            },
        )

        if block["data"].get("usingCaption"):
            caption = block["data"].get("alt")
            w = EditorJSElement(
                "figure",
                close_tag=True,
            )
            figcaption = EditorJSElement(
                "figcaption",
                caption,
            )
            w.append(imgTag)
            w.append(figcaption)
        else:
            w = imgTag

        return wrapper(w, attrs=attrs)
    
    @classmethod
    def get_test_data(cls):
        instance = Image.objects.first()
        return [
            {
                "imageId": instance.pk,
                "withBorder": True,
                "stretched": False,
                "backgroundColor": "#000000",
                "usingCaption": False,
                "alt": "Image",
                "caption": "Image",
            },
            {
                "imageId": instance.pk,
                "withBorder": False,
                "stretched": True,
                "backgroundColor": None,
                "usingCaption": True,
                "alt": "Image",
                "caption": "Image",
            }
        ]


class ImageRowFeature(BaseImageFeature):
    allowed_tags = ["div", "img"]
    allowed_attributes = ["class", "style"]
    klass = "ImageRowTool"
    js = [
        "wagtail_editorjs/js/tools/wagtail-image-row.js",
    ]

    def get_config(self, context: dict[str, Any]):
        config = super().get_config() or {}
        config.setdefault("config", {})
        config["config"]["imageChooserId"] =\
            f"editorjs-images-chooser-{context['widget']['attrs']['id']}"
        config["config"]["getImageUrl"] = reverse(f"wagtail_editorjs:{self.tool_name}_for_id_fmt")
        return config
    
    def render_template(self, context: dict[str, Any] = None):
        widget_id = f"editorjs-images-chooser-{context['widget']['attrs']['id']}"
        return AdminImageChooser().render_html(
            widget_id,
            None,
            {"id": widget_id}
        )

    def validate(self, data: Any):
        super().validate(data)

        if "images" not in data["data"]:
            raise forms.ValidationError("Invalid images value")
        
        if not data["data"]["images"]:
            raise forms.ValidationError("Invalid images value")
        
        if "settings" not in data["data"] or data["data"]["settings"] is None:
            raise forms.ValidationError("Invalid settings value")
        
        for image in data["data"]["images"]:
            if "id" not in image:
                raise forms.ValidationError("Invalid id value")
            
            if "title" not in image:
                raise forms.ValidationError("Invalid title value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        images = block["data"]["images"]
        ids = []
        for image in images:
            ids.append(image["id"])

        images = Image.objects.in_bulk(ids)
        s = []
        for id in ids:
            try:
                id = int(id)
            except ValueError:
                pass
            image = images[id]
            url = image.file.url
            if not any([url.startswith(i) for i in ["http://", "https://", "//"]]) and context:
                request = context.get("request")
                if request:
                    url = request.build_absolute_uri(url)

            s.append(EditorJSElement(
                "div",
                EditorJSElement(
                    "img",
                    close_tag=False,
                    attrs={
                        "src": url,
                        "alt": image.title,
                    }
                ),
                attrs={
                    "class": "image-wrapper",
                }
            ))


        return wrapper(
            EditorJSElement("div", s, attrs={
                "class": "image-row",
            }),
            attrs={
                "class": "stretched"
            } if block["data"]["settings"].get("stretched") else {}
        )

    @classmethod
    def get_test_data(cls):
        images = Image.objects.all()[0:3]
        return [
            {
                "images": [
                    {
                        "id": image.pk,
                        "title": image.title,
                    } for image in images
                ],
                "settings": {
                    "stretched": True,
                },
            }
        ]



from django.test import TestCase


from .utils import BaseEditorJSTest
from ..registry import (
    InlineEditorJSFeature,
    EDITOR_JS_FEATURES,
)

import bs4



TESTING_HTML = """
<div class="editorjs">
    <div class="editorjs-block" data-type="paragraph">
        <p>Hello, World!</p>
    </div>
    <div class="editorjs-block" data-type="header">
        <h1>Header</h1>
    </div>
    <div class="editorjs-block" data-type="list" data-testing-id="TARGET">

    </div>
    <div class="editorjs-block" data-type="table">
        <table>
            <tr>
                <td>1</td>
                <td>2</td>
            </tr>
            <tr>
                <td>3</td>
                <td>4</td>
            </tr>
        </table>
    </div>
</div>
"""


class TestEditorJSInline(BaseEditorJSTest):
    def setUp(self):
        super().setUp()
        self.inlines = [
            feature
            for feature in EDITOR_JS_FEATURES.features.values()
            if isinstance(feature, InlineEditorJSFeature)
        ]

    def test_inlines(self):

        for feature in self.inlines:
            feature: InlineEditorJSFeature
            test_data = feature.get_test_data()

            if not test_data:
                continue

            soup = bs4.BeautifulSoup(TESTING_HTML, "html.parser")
            testing_block = soup.find("div", {"data-testing-id": "TARGET"})
            testing_block.clear()

            for i, (initial, _) in enumerate(test_data):
                initial_soup = bs4.BeautifulSoup(initial, "html.parser")
                initial_soup.attrs["data-testing-id"] = f"test_{i}"
                testing_block.append(initial_soup)

            feature.parse_inline_data(soup)

            html = str(soup)

            outputs = [i[1] for i in test_data]
            for i, output in enumerate(outputs):
                self.assertInHTML(
                    output,
                    html,
                )



from typing import Any
from django import forms
from django.utils.translation import gettext_lazy as _
from ..registry import (
    EditorJSFeature,
    EditorJSBlock,
    EditorJSElement,
    wrap_tag,
)


class NestedListElement(EditorJSElement):
    def __init__(self, tag: str, items: list[EditorJSElement], close_tag: bool = True, attrs: dict[str, Any] = None):
        super().__init__(tag=tag, content=None, close_tag=close_tag, attrs=attrs)
        self.items = items

    def __str__(self):
        return wrap_tag(self.tag, self.attrs, "".join([str(item) for item in self.items]), self.close_tag)
    
    @property
    def content(self):
        return "".join([str(item) for item in self.items])
    
    @content.setter
    def content(self, value):
        if isinstance(value, list):
            self.items = value
        else:
            self.items = [value]

    def append(self, item: "NestedListElement"):
        self.items.append(item)


def parse_list(items: list[dict[str, Any]], element: str, depth = 0) -> NestedListElement:
    s = []

    for item in items:
        content = item.get("content")
        items = item.get("items")
        s.append(f"<li>{content}")
        if items:
            s.append(parse_list(items, element, depth + 1))
        s.append(f"</li>")

    return NestedListElement(element, s, attrs={"class": "nested-list", "style": f"--depth: {depth}"})

class NestedListFeature(EditorJSFeature):
    allowed_tags = ["ul", "ol", "li"]
    allowed_attributes = ["class", "style"]
    klass="NestedList"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/nested-list.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        items = data["data"].get("items")
        if not items:
            raise forms.ValidationError("Invalid items value")
        
        if "style" not in data["data"]:
            raise forms.ValidationError("Invalid style value")
        
        if data["data"]["style"] not in ["ordered", "unordered"]:
            raise forms.ValidationError("Invalid style value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        element = "ol" if block["data"]["style"] == "ordered" else "ul"
        return parse_list(block["data"]["items"], element)

    @classmethod
    def get_test_data(cls):
        return [
            {
                "style": "unordered",
                "items": [
                    {
                        "content": "Item 1",
                        "items": [
                            {
                                "content": "Item 1.1",
                                "items": [
                                    {
                                        "content": "Item 1.1.1",
                                        "items": [],
                                    },
                                    {
                                        "content": "Item 1.1.2",
                                        "items": [],
                                    },
                                ],
                            },
                            {
                                "content": "Item 1.2",
                                "items": [],
                            },
                        ],
                    },
                    {
                        "content": "Item 2",
                        "items": [],
                    },
                ],
            },
        ]


class CheckListFeature(EditorJSFeature):
    allowed_tags = ["ul", "li"]
    allowed_attributes = ["class"]
    klass="Checklist"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/checklist.js",
    ]

    def validate(self, data: Any):
        super().validate(data)
        
        items = data["data"].get("items")
        if not items:
            raise forms.ValidationError("Invalid items value")
        
        for item in items:
            if "checked" not in item:
                raise forms.ValidationError("Invalid checked value")
            
            if "text" not in item:
                raise forms.ValidationError("Invalid text value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        s = []
        for item in block["data"]["items"]:
            class_ = "checklist-item"
            if item["checked"]:
                class_ += " checked"

            s.append(wrap_tag("li", {"class": class_}, item["text"]))

        return EditorJSElement("ul", "".join(s), attrs={"class": "checklist"})
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "items": [
                    {
                        "checked": True,
                        "text": "Item 1",
                    },
                    {
                        "checked": False,
                        "text": "Item 2",
                    },
                ],
            }
        ]




from django.conf import settings as django_settings


"""
Clean the HTML output of the EditorJS field using bleach.
This will remove any tags or attributes not allowed by the EditorJS features.
If you want to disable this, set it to False.

Optionally; cleaning can be FORCED by passing `clean=True` to the `render_editorjs_html` function.
"""
CLEAN_HTML = getattr(django_settings, 'EDITORJS_CLEAN_HTML', True)

"""
Add a block ID to each editorJS block when rendering.
This is useful for targeting the block with JavaScript,
or possibly creating some link from frontend to admin area.
"""
ADD_BLOCK_ID = getattr(django_settings, 'EDITORJS_ADD_BLOCK_ID', True)

"""
The attribute name to use for the block ID.
This is only used if `ADD_BLOCK_ID` is True.
"""
BLOCK_ID_ATTR = getattr(django_settings, 'EDITORJS_BLOCK_ID_ATTR', 'data-editorjs-block-id')




from typing import Any
from .inlines import (
    ModelInlineEditorJSFeature,
)
from wagtail.snippets.widgets import AdminSnippetChooser


class SnippetChooserModel:
    """ Utility class for type annotations """

    def build_element(self, soup_elem, context = None): ...


class BaseInlineSnippetChooserFeature(ModelInlineEditorJSFeature):
    model: SnippetChooserModel = None
    widget = AdminSnippetChooser

    def build_element(self, soup_elem, obj: SnippetChooserModel, context: dict[str, Any] = None):
        return obj.build_element(soup_elem, context)







from .feature_registry import (
    EditorJSFeatures,
)
from .features import (
    BaseEditorJSFeature,
    EditorJSFeature,
    EditorJSJavascriptFeature,
    EditorJSTune,
    FeatureViewMixin,
    InlineEditorJSFeature,
    ModelInlineEditorJSFeature,
    TemplateNotSpecifiedError,
)
from .value import (
    EditorJSBlock,
    EditorJSValue,
)
from .element import (
    EditorJSElement,
    EditorJSWrapper,
    wrapper,
    EditorJSElementAttribute,
    EditorJSStyleAttribute,
    wrap_tag,
    add_attributes,
)

def get_features(features: list[str] = None):
    if not features:
        features = list(EDITOR_JS_FEATURES.keys())

    for feature in features:
        if feature not in EDITOR_JS_FEATURES:
            raise ValueError(f"Unknown feature: {feature}")

    return features

EDITOR_JS_FEATURES = EditorJSFeatures()




from typing import Any, TYPE_CHECKING
from django import forms
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.utils.safestring import mark_safe
from django.http import (
    JsonResponse,
)

from wagtail.models import Collection
from wagtail.documents import (
    get_document_model,
)
from wagtail.documents.forms import (
    get_document_form,
)
if TYPE_CHECKING:
    from wagtail.documents.models import AbstractDocument

from ..registry import (
    EditorJSFeature,
    EditorJSBlock,
    EditorJSElement,
    FeatureViewMixin,
)

BYTE_SIZE_STEPS = [_("Bytes"), _("Kilobytes"), _("Megabytes"), _("Gigabytes"), _("Terabytes")]

def filesize_to_human_readable(size: int) -> str:
    for unit in BYTE_SIZE_STEPS:
        if size < 1024:
            break
        size /= 1024
    return f"{size:.0f} {unit}"



Document = get_document_model()
DocumentForm = get_document_form(Document)


class AttachesFeature(FeatureViewMixin, EditorJSFeature):
    allowed_tags = [
        "div", "p", "span", "a",
        "svg", "path",
    ]
    allowed_attributes = {
        "div": ["class"],
        "p": ["class"],
        "span": ["class"],
        "a": ["class", "href", "title"],
        "svg": ["xmlns", "width", "height", "fill", "class", "viewBox"],
        "path": ["d"],
    }
    klass="CSRFAttachesTool"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/attaches.js",
        "wagtail_editorjs/js/tools/attaches.js",
    ],


    def get_config(self, context: dict[str, Any] = None) -> dict:
        config = super().get_config(context)
        config.setdefault("config", {})
        config["config"]["endpoint"] = reverse(f"wagtail_editorjs:{self.tool_name}")
        return config


    def validate(self, data: Any):
        super().validate(data)

        if "file" not in data["data"]:
            raise forms.ValidationError("Invalid file value")
        
        if "id" not in data["data"]["file"] and not data["data"]["file"]["id"] and "url" not in data["data"]["file"]:
            raise forms.ValidationError("Invalid id/url value")
        
        if "title" not in data["data"]:
            raise forms.ValidationError("Invalid title value")
        

    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:

        document_id = block["data"]["file"]["id"]
        document = Document.objects.get(pk=document_id)
        url = document.url

        if not any([url.startswith(i) for i in ["http://", "https://", "//"]]) and context:
            request = context.get("request")
            if request:
                url = request.build_absolute_uri(url)

        if block["data"]["title"]:
            title = block["data"]["title"]
        else:
            if document:
                title = document.title
            else:
                title = url

        return EditorJSElement(
            "div",
            [
                EditorJSElement(
                    "p",
                    EditorJSElement(
                        "a",
                        title,
                        attrs={"href": url},
                    ),
                    attrs={"class": "attaches-title"},
                ),
                EditorJSElement(
                    "span",
                    filesize_to_human_readable(document.file.size),
                    attrs={"class": "attaches-size"},
                ),
                EditorJSElement(
                    "a",
                    mark_safe("""<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-link" viewBox="0 0 16 16">
    <path d="M6.354 5.5H4a3 3 0 0 0 0 6h3a3 3 0 0 0 2.83-4H9q-.13 0-.25.031A2 2 0 0 1 7 10.5H4a2 2 0 1 1 0-4h1.535c.218-.376.495-.714.82-1z"/>
    <path d="M9 5.5a3 3 0 0 0-2.83 4h1.098A2 2 0 0 1 9 6.5h3a2 2 0 1 1 0 4h-1.535a4 4 0 0 1-.82 1H12a3 3 0 1 0 0-6z"/>
</svg>"""),
                    attrs={
                        "title": _("Download"),
                        "href": url,
                        "class": "attaches-link",
                        # "data-id": document_id,
                    },
                )
            ],
            attrs={"class": "attaches"},
        )
    

    @classmethod
    def get_test_data(cls):
        instance = Document.objects.first()
        return [
            {
                "file": {
                    "id": instance.pk,
                },
                "title": "Document",
            },
        ]

    @csrf_exempt
    def handle_post(self, request):
        file = request.FILES.get('file')
        if not file:
            return JsonResponse({
                'success': False,
                'errors': {
                    'file': ["This field is required."]
                }
            }, status=400)

        filename = file.name
        title = request.POST.get('title', filename)

        collection = Collection.get_first_root_node().id
        form = DocumentForm({ 'title': title, 'collection': collection }, request.FILES)
        if form.is_valid():
            document: AbstractDocument = form.save(commit=False)

            hash = document.get_file_hash()
            existing = Document.objects.filter(file_hash=hash)
            if existing.exists():
                exists: AbstractDocument = existing.first()
                return JsonResponse({
                    'success': True,
                    'file': {
                        'id': exists.pk,
                        'title': exists.title,
                        'size': exists.file.size,
                        'url': exists.url,
                        'upload_replaced': True,
                        'reuploaded_by_user': request.user.pk,
                    }
                })

            document.uploaded_by_user = request.user
            document.save()
            return JsonResponse({
                'success': True,
                'file': {
                    'id': document.pk,
                    'title': document.title,
                    'size': document.file.size,
                    'url': document.url,
                    'upload_replaced': False,
                    'reuploaded_by_user': None,
                }
            })
        else:
            return JsonResponse({
                'success': False,
                'errors': form.errors,
            }, status=400)




from typing import Any, Union
from collections import defaultdict
from django.template.loader import render_to_string
from django.template.context import Context
from django.utils.safestring import mark_safe
from . import settings
from .registry import (
    EditorJSElement,
    InlineEditorJSFeature,
    EDITOR_JS_FEATURES,
)
import bleach, bs4


class NullSanitizer:
    @staticmethod
    def sanitize_css(val):
        return val

def render_editorjs_html(
        features: list[str],
        data: dict,
        context=None,
        clean: bool = None,
        whitelist_tags: list[str] = None,
        whitelist_attrs: Union[dict, list] = None
    ) -> str:
    """
        Renders the editorjs widget based on the features provided.
    """

    if "blocks" not in data:
        data["blocks"] = []

    feature_mappings = {
        feature: EDITOR_JS_FEATURES[feature]
        for feature in features
    }

    inlines = [
        feature
        for feature in feature_mappings.values()
        if isinstance(feature, InlineEditorJSFeature)
    ]

    html = []
    for block in data["blocks"]:

        feature: str = block["type"]
        tunes: dict[str, Any] = block.get("tunes", {})
        feature_mapping = feature_mappings.get(feature, None)

        if not feature_mapping:
            continue

        # Build the actual block.
        element: EditorJSElement = feature_mapping.render_block_data(block, context)

        # Optionally tools can decide to not render the block.
        if element is None:
            continue

        # Tune the element.
        for tune_name, tune_value in tunes.items():
            element = feature_mappings[tune_name].tune_element(element, tune_value, context)

        # Add the block ID to each individual block.
        if settings.ADD_BLOCK_ID:
            # This can be used to link frontend to the admin area.
            element.attrs[settings.BLOCK_ID_ATTR] = block.get("id", "")
        
        html.append(element)

    html = "\n".join([str(h) for h in html])

    soup = bs4.BeautifulSoup(html, "html.parser")
    if inlines:
        for inline in inlines:
            # Give inlines access to whole soup.
            # This allows for proper parsing of say; page or document links.
            inline: InlineEditorJSFeature
            inline.parse_inline_data(soup, context)

        # Re-render the soup.
        html = soup.decode(False)

    if clean or (clean is None and settings.CLEAN_HTML):
        allowed_tags = set({
            # Default inline tags.
            "i", "b", "strong", "em", "u", "s", "strike"
        })
        allowed_attributes = defaultdict(set)
        # cleaner_funcs = defaultdict(lambda: defaultdict(list))

        for feature in feature_mappings.values():
            allowed_tags.update(feature.allowed_tags)
            # for key, value in feature.cleaner_funcs.items():
            #     for name, func in value.items():
            #         cleaner_funcs[key][name].append(func)

            for key, value in feature.allowed_attributes.items():
                allowed_attributes[key].update(value)

        if whitelist_tags:
            allowed_tags.update(whitelist_tags)

        if "*" in allowed_attributes:
            allowed_attributes["*"].add(settings.BLOCK_ID_ATTR)
        else:
            allowed_attributes["*"] = {settings.BLOCK_ID_ATTR}

        if whitelist_attrs:
            if isinstance(whitelist_attrs, dict):
                for key, value in whitelist_attrs.items():
                    allowed_attributes[key].update(value)
            else:
                for key in allowed_attributes:
                    allowed_attributes[key].update(whitelist_attrs)

        html = bleach.clean(
            html,
            tags=allowed_tags,
            attributes=allowed_attributes,
            css_sanitizer=NullSanitizer,
        )

    ctx = context or {}
    ctx["html"] = html

    if isinstance(context, Context):
        ctx = context.flatten()

    return render_to_string(
        "wagtail_editorjs/rich_text.html", ctx
    )



#         def parse_allowed_attributes(tag, name, value):
#             if (
#                 tag not in allowed_attributes\
#                 and tag not in cleaner_funcs\
#                 and "*" not in cleaner_funcs\
#                 and "*" not in allowed_attributes
#             ):
#                 return False
#             
#             if "*" in cleaner_funcs and name in cleaner_funcs["*"] and any(
#                 func(value) for func in cleaner_funcs["*"][name]
#             ):
#                 return True
#             
#             if tag in cleaner_funcs\
#                     and name in cleaner_funcs[tag]\
#                     and any(
#                         func(value) for func in cleaner_funcs[tag][name]
#                     ):
#                 return True
#             
#             if name in allowed_attributes[tag] or name in allowed_attributes["*"]:
#                 return True
#             
#             return False



from typing import Any
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from wagtail.models import Page
from wagtail.admin.widgets import AdminPageChooser
from wagtail.documents.widgets import AdminDocumentChooser
from wagtail.documents import get_document_model
from wagtail import hooks

from ..registry import (
    ModelInlineEditorJSFeature,
    FeatureViewMixin,
)
from django.http import (
    HttpResponse,
    JsonResponse,
)


Document = get_document_model()


class BasePageLinkMixin:
    @classmethod
    def get_url(cls, instance):
        return instance.get_url()

    @classmethod
    def get_full_url(cls, instance, request):
        return instance.get_full_url(request)
    
    @classmethod
    def get_test_queryset(cls):
        return super().get_test_queryset().filter(depth__gt=1)

class LinkFeature(BasePageLinkMixin, ModelInlineEditorJSFeature):
    allowed_tags = ["a"]
    allowed_attributes = ["class", "href", "data-id"]
    must_have_attrs = {
        "data-parent-id": None,
    }
    chooser_class = AdminPageChooser
    model = Page

    klass="WagtailLinkTool"
    js = [
        "wagtail_editorjs/js/tools/wagtail-chooser-tool.js",
        "wagtail_editorjs/js/tools/wagtail-link.js",
    ]

    @classmethod
    def get_test_data(cls):
        models = cls.get_test_queryset()[0:5]
        return [
            (
                # Override to add data-autocomplete.
                f"<a data-id='{model.id}' data-{cls.model._meta.model_name}='True' data-parent-id='this-doesnt-get-used'></a>",
                f"<a href='{cls.get_url(model)}' class='{cls.model._meta.model_name}-link'></a>",
            )
            for model in models
        ]


SEARCH_QUERY_PARAM = "search"
CONSTRUCT_PAGE_QUERYSET = "construct_page_queryset"
BUILD_PAGE_DATA = "build_page_data"


class LinkAutoCompleteFeature(FeatureViewMixin, BasePageLinkMixin, ModelInlineEditorJSFeature):
    allowed_tags = ["a"]
    allowed_attributes = ["class", "href", "data-id"]
    must_have_attrs = {
        "data-autocomplete": "page",
    }
    chooser_class = AdminPageChooser
    model = Page
    klass="LinkAutocomplete"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/link-autocomplete.js",
    ]

    def get_config(self, context: dict[str, Any]):
        config = super(ModelInlineEditorJSFeature, self).get_config(context)
        config.setdefault("config", {})
        config["config"]["endpoint"] = reverse(f"wagtail_editorjs:{self.tool_name}")
        config["config"]["queryParam"] = "search"
        return config

    @classmethod
    def get_test_data(cls):
        models = cls.get_test_queryset()[0:5]
        return [
            (
                # Override to add data-autocomplete.
                f"<a data-id='{model.id}' data-{cls.model._meta.model_name}='True' data-autocomplete='page'></a>",
                f"<a href='{cls.get_url(model)}' class='{cls.model._meta.model_name}-link'></a>",
            )
            for model in models
        ]

    def handle_get(self, request):
        """
        Autocomplete for internal links
        """
        search = request.GET.get(SEARCH_QUERY_PARAM)

        pages = Page.objects.all()\
            .live()\
            .specific()

        for fn in hooks.get_hooks(CONSTRUCT_PAGE_QUERYSET):
            pages = fn(request, pages, search)

            if isinstance(pages, HttpResponse):
                return pages

        if search:
            pages = pages.search(search)


        page_list = []
        page_data_hooks = hooks.get_hooks(BUILD_PAGE_DATA)
        for page in pages:
            data = {
                'id': page.id,
                'name': page.title,
                'href': page.get_url(request),
                'description': page.search_description,
                'autocomplete': 'page',
                'page': True,
            }

            for hook in page_data_hooks:
                hook(request, page, data)

            page_list.append(data)


        return JsonResponse({
                "success": True,
                "items": page_list,
            },
            status=200,
        )


class DocumentFeature(ModelInlineEditorJSFeature):
    allowed_tags = ["a"]
    allowed_attributes = ["class", "href", "data-id"]
    chooser_class = AdminDocumentChooser
    model = Document
    klass = "WagtailDocumentTool"
    js = [
        "wagtail_editorjs/js/tools/wagtail-chooser-tool.js",
        "wagtail_editorjs/js/tools/wagtail-document.js",
    ]

    @classmethod
    def get_url(cls, instance):
        return instance.file.url




from typing import Any
from wagtail import hooks
from django.urls import reverse_lazy

from ..hooks import (
    REGISTER_HOOK_NAME,
    BUILD_CONFIG_HOOK,
)
from ..registry import (
    EditorJSFeature,
    EditorJSFeatures,
    EditorJSJavascriptFeature,
)
from ..features import (
    AttachesFeature,
    BlockQuoteFeature,
    CheckListFeature,
    CodeFeature,
    DelimiterFeature,
    WarningFeature,
    HeaderFeature,
    HTMLFeature,
    ImageFeature,
    ImageRowFeature,
    LinkFeature,
    LinkAutoCompleteFeature,
    DocumentFeature,
    NestedListFeature,
    TableFeature,
    AlignmentBlockTune,
    TextVariantTune,
    BackgroundColorTune,
    ColorTune,
)

@hooks.register(BUILD_CONFIG_HOOK)
def build_editorjs_config(widget, context, config):
    config["minHeight"] = 150


@hooks.register(REGISTER_HOOK_NAME)
def register_editor_js_features(registry: EditorJSFeatures):

    registry.register(
        "attaches",
        AttachesFeature(
            "attaches",
        ),
    )
    registry.register(
        "checklist",
        CheckListFeature(
            "checklist",
            inlineToolbar = True,
        
        ),
    )
    registry.register(
        "code",
        CodeFeature(
            "code",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "delimiter",
        DelimiterFeature(
            "delimiter",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "header",
        HeaderFeature(
            "header",
            inlineToolbar = True,
        
        ),
    )
    registry.register(
        "inline-code",
        EditorJSFeature(
            "inline-code",
            "InlineCode",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/inline-code.js",
            ],
            inlineToolbar = True,
        ),
    )
    registry.register(
        "marker",
        EditorJSFeature(
            "marker",
            "Marker",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/marker.js",
            ],
            inlineToolbar = True,
            allowed_tags=["mark"],
        ),
    )
    registry.register(
        "nested-list",
        NestedListFeature(
            "nested-list",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "paragraph",
        EditorJSFeature(
            "paragraph",
            "Paragraph",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/paragraph.umd.js",
            ],
            inlineToolbar = True,
            allowed_tags=["p"],
        ),
    )
    registry.register(
        "quote",
        BlockQuoteFeature(
            "quote",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "raw",
        HTMLFeature(
            "raw",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "table",
        TableFeature(
            "table",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "underline",
        EditorJSFeature(
            "underline",
            "Underline",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/underline.js",
            ],
            inlineToolbar = True,
        ),
    )
    registry.register(
        "warning",
        WarningFeature(
            "warning",
            inlineToolbar = True,
        ),
    )
    
    registry.register(
        "link-autocomplete",
        LinkAutoCompleteFeature(
            "link-autocomplete",
        ),
    )

    # Wagtail specific
    registry.register(
        "link",
        LinkFeature(
            "link",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "image",
        ImageFeature(
            "image",
        ),
    )
    registry.register(
        "images",
        ImageRowFeature(
            "images",
        ),
    )
    registry.register(
        "document",
        DocumentFeature(
            "document",
            config={},
            inlineToolbar = True,
        ),
    )

    # Tunes
    registry.register(
        "text-alignment-tune",
        AlignmentBlockTune(
            "text-alignment-tune",
            inlineToolbar = True,
            config={
                "default": "left",
            },
        ),
    )
    registry.register(
        "text-variant-tune",
        TextVariantTune(
            "text-variant-tune",
            inlineToolbar = True,
        ),
    )
    registry.register(
        "background-color-tune",
        BackgroundColorTune(
            "background-color-tune",
            config={
                "defaultStretched": True,
            }
        ),
    )
    registry.register(
        "text-color-tune",
        ColorTune(
            "text-color-tune",
        ),
    )

    # Register initializers
    registry.register(
        "undo-redo",
        EditorJSJavascriptFeature(
            "undo-redo",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/undo-redo.js",
                "wagtail_editorjs/js/init/undo-redo.js",
            ],
            weight=0,
        ),
    )

    registry.register(
        "drag-drop",
        EditorJSJavascriptFeature(
            "drag-drop",
            js = [
                "wagtail_editorjs/vendor/editorjs/tools/drag-drop.js",
                "wagtail_editorjs/js/init/drag-drop.js",
            ],
            weight=1,
        ),
    )

    # Add tunes
    registry.register_tune("text-alignment-tune")
    registry.register_tune("text-variant-tune")
    registry.register_tune("background-color-tune")
    registry.register_tune("text-color-tune")


# 
# from django.contrib.auth import get_user_model
# from wagtail.snippets.widgets import AdminSnippetChooser
# from wagtail_editorjs.hooks import REGISTER_HOOK_NAME
# from wagtail_editorjs.registry import (
#     EditorJSFeatures,
#     ModelInlineEditorJSFeature,
# )
# 
# 
# class InlineUserChooserFeature(ModelInlineEditorJSFeature):
#     model = get_user_model()
#     chooser_class = AdminSnippetChooser
#     must_have_attrs = {
#         "data-email": None,
#     }
# 
#     @classmethod
#     def get_url(cls, instance):
#         return f"mailto:{instance.email}"
#     
#     @classmethod
#     def get_full_url(cls, instance, request):
#         return cls.get_url(instance)
# 
# 
# @hooks.register(REGISTER_HOOK_NAME)
# def register_editor_js_features(registry: EditorJSFeatures):
#     registry.register(
#         "inline-user",
#         InlineUserChooserFeature(
#             "inline-user",
#             "InlineUserTool",
#             js=[
#                 "wagtail_editorjs/js/tools/inline-user.js",
#             ],
#             config={},
#         ),
#     )
# 


"""
wagtail_editorjs
© 2024 Nigel van Keulen, goodadvice.it

A Wagtail integration for the Editor.js block editor.
Includes support for custom blocks, inline tools, and more.
Everything is dynamic and customizable.
"""

__version__ = '1.5.6'
VERSION = tuple([int(v) for v in __version__.split('.')])



from typing import Any
from django import forms
from django.utils.translation import gettext_lazy as _

import bleach

from ..registry import (
    EditorJSFeature,
    EditorJSBlock,
    EditorJSElement,
    wrap_tag,
)



class CodeFeature(EditorJSFeature):
    allowed_tags = ["code"]
    allowed_attributes = ["class"]
    klass="CodeTool"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/code.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if 'code' not in data['data']:
            raise forms.ValidationError('Invalid code value')
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("code", block["data"]["code"], attrs={"class": "code"})
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "code": "print('Hello, World!')",
            }
        ]


class DelimiterFeature(EditorJSFeature):
    allowed_tags = ["hr"]
    allowed_attributes = ["class"]
    klass="Delimiter"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/delimiter.js",
    ]

    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("hr", close_tag=False, attrs={"class": "delimiter"})
    
    @classmethod
    def get_test_data(cls):
        return [{}, {}, {}]


class HeaderFeature(EditorJSFeature):
    allowed_tags = ["h1", "h2", "h3", "h4", "h5", "h6"]
    allowed_attributes = ["class"]
    klass="Header"
    js=[
        "wagtail_editorjs/vendor/editorjs/tools/header.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        level = data["data"].get("level")
        if level > 6 or level < 1:
            raise forms.ValidationError("Invalid level value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement(
            "h" + str(block["data"]["level"]),
            block["data"].get("text")
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "level": heading,
                "text": f"Header {heading}",
            } for heading in range(1, 7)
        ]


class HTMLFeature(EditorJSFeature):
    allowed_tags = bleach.ALLOWED_TAGS
    allowed_attributes = bleach.ALLOWED_ATTRIBUTES
    klass="RawTool"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/raw.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "html" not in data["data"]:
            raise forms.ValidationError("Invalid html value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement("div", block["data"]["html"], attrs={"class": "html"})
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "html": "<p>This is a HTML block.</p>",
            }
        ]

class WarningFeature(EditorJSFeature):
    allowed_tags = ["div", "h2", "p"]
    allowed_attributes = ["class"]
    klass="Warning"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/warning.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "title" not in data["data"]:
            raise forms.ValidationError("Invalid title value")
        
        if "message" not in data["data"]:
            raise forms.ValidationError("Invalid message value")
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        return EditorJSElement(
            "div",
            attrs={
                "class": "warning",
            },
            content=[
                EditorJSElement(
                    "h2",
                    block["data"]["title"],
                ),
                EditorJSElement(
                    "p",
                    block["data"]["message"],
                ),
            ]
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "title": "Warning",
                "message": "This is a warning message.",
            }
        ]


class TableFeature(EditorJSFeature):
    allowed_tags = ["table", "tr", "th", "td", "thead", "tbody", "tfoot"]
    allowed_attributes = ["class"]
    klass="Table"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/table.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "content" not in data["data"]:
            raise forms.ValidationError("Invalid content value")
        
        if "withHeadings" not in data["data"]:
            raise forms.ValidationError("Invalid withHeadings value")

    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        table = []

        if block["data"]["withHeadings"]:
            headings = block["data"]["content"][0]
            table.append(EditorJSElement(
                "thead",
                EditorJSElement(
                    "tr",
                    [
                        EditorJSElement("th", heading)
                        for heading in headings
                    ]
                )
            ))

            content = block["data"]["content"][1:]
        else:
            content = block["data"]["content"]

        tbody = EditorJSElement("tbody")
        for row in content:
            tbody.append(
                EditorJSElement(
                    "tr",
                    [
                        EditorJSElement("td", cell)
                        for cell in row
                    ]
                )
            )

        table.append(tbody)

        return EditorJSElement("table", table)

    @classmethod
    def get_test_data(cls):
        return [
            {
                "withHeadings": False,
                "content": [
                    ["1", "2", "3"],
                    ["4", "5", "6"],
                    ["7", "8", "9"],
                ],
            },
            {
                "withHeadings": True,
                "content": [
                    ["Heading 1", "Heading 2", "Heading 3"],
                    ["1", "2", "3"],
                    ["4", "5", "6"],
                    ["7", "8", "9"],
                ],
            }
        ]


class BlockQuoteFeature(EditorJSFeature):
    allowed_tags = ["blockquote", "footer"]
    allowed_attributes = ["class"]
    klass="Quote"
    js = [
        "wagtail_editorjs/vendor/editorjs/tools/quote.js",
    ]

    def validate(self, data: Any):
        super().validate(data)

        if "text" not in data["data"]:
            raise forms.ValidationError("Invalid text value")
        
        if "caption" not in data["data"]:
            raise forms.ValidationError("Invalid caption value")
        
    
    def render_block_data(self, block: EditorJSBlock, context = None) -> EditorJSElement:
        text = block["data"]["text"]
        caption = block["data"]["caption"]
        return EditorJSElement(
            "blockquote",
            [
                text,
                wrap_tag("footer", {}, caption),
            ],
            {
                "class": "blockquote",
            }
        )
    
    @classmethod
    def get_test_data(cls):
        return [
            {
                "text": "This is a quote.",
                "caption": "Anonymous",
            }
        ]




from .base import (
    TemplateNotSpecifiedError,
    BaseEditorJSFeature,
)
from .basic import (
    EditorJSFeature,
    EditorJSJavascriptFeature,
    EditorJSTune,
)
from .inlines import (
    InlineEditorJSFeature,
    ModelInlineEditorJSFeature,
)
from .view import (
    FeatureViewMixin,
)


"""
    This module defines mappings from tools to the editorjs javascript side.
"""

from collections import defaultdict, OrderedDict
from typing import Any, Union
from wagtail import hooks

import copy
from ..hooks import (
    REGISTER_HOOK_NAME,
)
from .value import (
    EditorJSValue,
)

from .features import (
    BaseEditorJSFeature,
    EditorJSFeature,
    EditorJSTune,
    InlineEditorJSFeature,
)


class EditorJSFeatures:
    def __init__(self):
        self.features: dict[str, Union[EditorJSFeature, EditorJSTune]] = {}
        self.inline_features: list[InlineEditorJSFeature] = []
        self.tunes_for_all: list[str] = []
        self.tunes_for_tools: defaultdict[str, list[str]] = defaultdict(list)
        self._looked_for_features = False


    def __contains__(self, tool_name: str):
        self._look_for_features()
        return tool_name in self.features


    def __getitem__(self, tool_name: str) -> Union[EditorJSFeature, EditorJSTune, InlineEditorJSFeature]:
        self._look_for_features()

        if isinstance(tool_name, BaseEditorJSFeature):
            return tool_name

        return self.features[tool_name]

    def keys(self):
        self._look_for_features()
        return self.features.keys()


    def _look_for_features(self):
        if not self._looked_for_features:
            for hook in hooks.get_hooks(REGISTER_HOOK_NAME):
                hook(self)
            self._looked_for_features = True


    def register(self, tool_name: str, feature: EditorJSFeature):
        """
            Register a feature with the EditorJS editor.
            After registering it; you can easily use this by
            providing the `tool_name` into `features` for widgets and fields.
        """
        self.features[tool_name] = feature
        if isinstance(feature, InlineEditorJSFeature):
            self.inline_features.append(feature)

        # Callback to the feature in case it needs to do something
        # after being registered. (like registering admin urls.)
        feature.on_register(self)


    def register_tune(self, tune_name: str, tool_name: str = None):
        """
            Register a tune (BY NAME) for a tool.
            This tune will be made available to only that tool (unless otherwise specified)
            If not providing a `tool_name`; the tune will be available to all tools.
        """
        if tool_name:
            self.tunes_for_tools[tool_name].append(tune_name)
        else:
            self.tunes_for_all.append(tune_name)


    def register_config(self, tool_name: str, config: dict):
        """
            Register or override any additional configuration for a tool.
        """
        self._look_for_features()
        self.features[tool_name].config.update(config)


    def build_config(self, tools: list[str], context: dict[str, Any] = None):
        """
            Builds out the configuration for the EditorJS widget.
            This config is passed into the editorjs javascript side.
        """
        editorjs_config = {}
        editorjs_config_tools = {}
        self._look_for_features()

        for tool in tools:
            if tool not in self.features:
                raise ValueError(f"Unknown feature: {tool}")

            tool_mapping = self.features[tool]
            tool_config = tool_mapping.get_config(context)
            if tool_config is None:
                continue

            tool_config = copy.deepcopy(tool_config)

            if "tunes" in tool_config:
                raise ValueError(f"Tunes must be registered separately for {tool}")

            tunes = self.tunes_for_tools[tool]
            if tunes:
                tool_config["tunes"] = tunes

            editorjs_config_tools[tool] = tool_config

        if self.tunes_for_all:
            editorjs_config["tunes"] = list(
                filter(lambda tune: tune in tools, self.tunes_for_all)
            )

        editorjs_config["tools"] = editorjs_config_tools

        return editorjs_config
    

    def to_python(self, tools: list[str], data: list[dict]):
        """
            Converts the data from the editorjs format to the python format.
        """

        if isinstance(data, EditorJSValue):
            return data

        self._look_for_features()
        block_list = data.get("blocks", [])
        features = {
            tool: self.features[tool]
            for tool in tools
        }

        for i, item in enumerate(block_list):
            block_type = item.get("type")
            if block_type not in tools:
                continue

            tool_mapping = features[block_type]
            block_list[i] = tool_mapping.create_block(tools, item)

        data["blocks"] = block_list

        return EditorJSValue(
            data,
            {
                tool: self.features[tool]
                for tool in tools
            }
        )


    def prepare_value(self, tools: list[str], data: dict):
        """
            Filters out unknown features and tunes.
            Return the value back to native format.
        """
        self._look_for_features()
        block_list = data.get("blocks", [])
        blocks = [None] * len(block_list)
        for i, item in enumerate(block_list):

            block_type = item.get("type")
            if block_type not in tools:
                continue

            tunes = item.get("tunes", {})
            if tunes:
                keys = list(tunes.keys())
                for key in keys:
                    if key not in tools:
                        del tunes[key]

            blocks[i] = item

        data["blocks"] = list(filter(None, blocks))
        return data

    def get_by_weight(self, tools: list[str]) -> OrderedDict[str, EditorJSFeature]:
        """
            Returns the tools sorted by weight.
            Items with the lowest weight are first.
        """
        self._look_for_features()
        od = OrderedDict()
        
        values = list(self.features.values())
        values.sort(key=lambda x: x.weight)
        
        for value in values:
            if value.tool_name in tools:
                od[value.tool_name] = value
        
        return od

    def validate_for_tools(self, tools: list[str], data: dict):
        """
            Validates the data for the given tools.
        """
        self._look_for_features()

        block_list = data.get("blocks", [])

        for tool in tools:
            if tool not in self.features:
                raise ValueError(f"Unknown feature: {tool}")
            
            tool_mapping = self.features[tool]

            for item in block_list:
                if isinstance(tool_mapping, EditorJSTune):
                    tunes = item.get("tunes", {})
                    tune = tunes.get(tool)
                    if tune:
                        tool_mapping.validate(tune)

                else:
                    if item["type"] == tool:
                        tool_mapping.validate(item)

        return data






from django.test import TestCase
from wagtail.models import Collection, Page
from wagtail.images.tests.utils import (
    get_test_image_file,
    get_test_image_file_jpeg,
    get_test_image_file_webp,
    get_test_image_file_avif,
)
from wagtail.images import get_image_model
from wagtail.documents import get_document_model


Image = get_image_model()
Document = get_document_model()


class BaseEditorJSTest(TestCase):
    """
        Setup test data for EditorJS tests
        This is so blocks can freely be tested and query the test db
        without having to worry about setting up the data.
    """
    def setUp(self) -> None:
        image_funcs = [
            get_test_image_file,
            get_test_image_file_jpeg,
            get_test_image_file_webp,
            get_test_image_file_avif,
        ]
        self.collection = Collection.get_first_root_node()
        root_page = Page.objects.filter(depth=2).first()
        
        child_url_paths = []
        sibling_url_paths = []
        
        for i in range(100):
            
            child_url_paths.append(f"test-page-{i}")
            sibling_url_paths.append(f"test-subchild-page-{i}")

            child = root_page.add_child(instance=Page(
                title=f"Test Page {i}",
                slug=f"test-page-{i}",
                path=f"/test-page-{i}/",
                url_path=f"/test-page-{i}/",
            ))
            child.set_url_path(root_page)
            child.save()

            subchild = child.add_child(instance=Page(
                title=f"Test Page subchild {i}",
                slug=f"test-subchild-page-{i}",
                path=f"/test-subchild-page-{i}/",
                url_path=f"/test-subchild-page-{i}/",
            ))
            subchild.set_url_path(child)
            subchild.save()

            Document.objects.create(file=get_test_image_file(), title=f"Test Document {i}", collection=self.collection)
            Image.objects.create(
                file=image_funcs[i % len(image_funcs)](),
                title=f"Test Image {i}",
                collection=self.collection
            )
        # call_command("fixtree")
        # call_command("set_url_paths")




from typing import Any, Union


class EditorJSElementAttribute:
    def __init__(self, value: Union[str, list[str]], delimiter: str = " "):
        if not isinstance(value, (list, dict)):
            value = [value]

        if not isinstance(value, dict):
            value = set(value)

        self.value = value
        self.delimiter = delimiter

    def __eq__(self, other):
        if isinstance(other, EditorJSElementAttribute):
            return self.value == other.value
        
        if isinstance(other, (tuple, list, set)):
            for item in other:
                if item not in self.value:
                    return False
            return True
        return False

    def extend(self, value: Any):

        if isinstance(value, (tuple, list, set)):
            self.value.update(value)

        else:
            self.value.add(value)

    def __str__(self):
        return self.delimiter.join([str(item) for item in self.value])
    

class EditorJSStyleAttribute(EditorJSElementAttribute):
    def __init__(self, value: dict):
        super().__init__(value, ";")

    def __repr__(self):
        return f"EditorJSStyleAttribute({self.value})"

    def __eq__(self, other):
        if isinstance(other, EditorJSStyleAttribute):
            return self.value == other.value
        
        if isinstance(other, dict):
            return self.value == other
        
        if isinstance(other, str):
            try:
                key, value = other.split(":")
                return self.value.get(key) == value.strip()
            except ValueError:
                return False
        return False

    def extend(self, value: dict = None, **kwargs):
        if value:
            if not isinstance(value, dict):
                raise ValueError("Value must be a dictionary")
            self.value.update(value)

        self.value.update(kwargs)

    def __str__(self):
        return self.delimiter.join([f'{key}: {value}' for key, value in self.value.items()])
